<%--
 * Component: dotcom
 *
 * Description:
 *   WEB interface to accept database synchronization requests.
 *
 * Parameters:
 *
 *		cStudy          - Study Instance UID of the study to be synchronized
 *		cForced					-	'yes' if it's a forced (full state) sending 
 *		cNumTables      - Number of database tables 
 *		cTableNamex     - Name of the x-th database table
 *		cTableRowsx     - Number of table entries in the x-th database table
 *		cUniqAttrx      - Name of the unique attribute (field) in the x-th database table
 *		cUniqValuex     - Values (not quoted - separated by newline) for the unique attribute in the x-th database table entries.
 *		cFieldsx        - Comma separated list of field names in the x-th database table
 *		cValuesx        - Database table entries (separated  by newline) in the c-th database table.
 *		                  Each line contains comma separated list of quoted values in the same order as field names.
 *		cAutoRetrieve   - Start autoretrieve
 *		cNumSources			- CWL to CWL only: Number of Sources as known by this CWL
 *
 *	Quoted Value:
 *	 - String values are SQL encoded and appears between single quotes
 *	 - Numeric values are as is
 *	 - NULL is represented by "null"
 *   
--%><%@
	page import="java.io.IOException, java.sql.*, java.util.*, imagemed.imsession.*, imagemed.exception.IMException,
			imagemed.util.*, imagemed.jsputil.*, imagemed.taskd.IMScriptContext, imagemed.dotcom.*"
	
%><%!
	static Syslog logger = new Syslog( "syncStudy.jsp", 0, IMEnvBase.getEnv( "LOG_FACILITY" ) );
	static Syslog operLogger = new Syslog( "syncStudy.jsp", 0, IMEnvBase.getEnv( "LOG_OPER_FACILITY" ) );

	static final String REFERENCE_COUNTER_COMMENT = "syncStudy-CW1";

	static final int DO_NOTHING = 0x00;
	static final int DO_SYNCDB  = 0x01;
	static final int DO_CHANGE_OWNER = 0x02;

	/**
	 * Strips starting and ending (single) quotes from value if any at all.
	 *
	 * @param value to be stripped
	 *
	 * @return  value with starting and ending (single) quotes removed 
	 */
	static String stripQuotes( String value ) {
		if ( value == null ) {
			return null;
		}
		if ( value.startsWith( "'" ) ) {
			value = value.substring( 1 );
		}
		if ( value.endsWith( "'" ) ) {
			value = value.substring( 0, value.length() - 1 );
		}
		
		return value;
	}
	
	
	/**
	 * Checks if the storestate.rec file is correct and adds/removes peer if necessary.
	 *
	 * @param peerId        Device ID of the peer
	 * @param storestate    content of the storestate.rec file of the study being synchronized
	 * @param isDeleted     true indicates that the study is supposed to be deleted on the peer
	 *
	 * @return true todo flags (combination of DO_NOTHING, DO_SYNCDB, DO_UPDATE_NUMSRC, DO_CHANGE_OWNER)
	 *
	 * @throws IMException if an error occurs during manipulating the storestate.rec file
	 */
	int checkAndFixStorestate( String peerId, StoreState storestate, boolean isDeleted ) throws IMException {
		String ownerDevice = storestate.getAttribute( "OwnerDevice", "NULL" );
		Set<String> stagedDevices = storestate.getStagedDevices();
		int returnValue = DO_NOTHING;

		Vector<String> modStorestate = new Vector<String>();

		if ( !"Database".equals( storestate.getAttribute( "ProcessMode" ) ) ) {
			modStorestate.add( "--processmode" );
			modStorestate.add( "Database" );
		}

		if ( peerId.equals( ownerDevice ) ) {
			returnValue |= DO_SYNCDB;
			if ( !isDeleted ) {
				// The peer has already been set as owner device and the study still exists on peer
				return returnValue;
			}

			// The peer is the owner but the study is deleted from the peer - the peer should be removed from storestate.rec
			if ( stagedDevices.isEmpty() ) {
				// but if there is no stage devices for the study, do not touch the storestate, it will be removed after the references are removed
				return returnValue;
			}

			// Remove peer as Owner Device - set one of the staged devices as owner device
			modStorestate.add( "--removedevice" );
			modStorestate.add( peerId );

			returnValue |= DO_CHANGE_OWNER;
			
			try {
				String newOwner = stagedDevices.iterator().next();
				// newOwner could not be null, but for safety
				if ( newOwner != null ) {
					modStorestate.add( "--setowner" );
					modStorestate.add( newOwner );
				}
			}
			catch ( Throwable th ) {
				// iterator().next() might throw NoSuchElementException - but it should never happen since set is not empty
			}	
		}
		else if ( "NULL".equals( ownerDevice ) ) {
			// No owner device set for the study
			if ( isDeleted ) {
				// The study is deleted from child - check if we are a staged device
				if ( stagedDevices.contains( peerId ) ) {
					// should not happen, but ...

					// remove peer from staged devices
					modStorestate.add( "--removedevice" );
					modStorestate.add( peerId );
				}
				else {
					return returnValue; //ie: DO_NOTHING
				}
			}

			// Set the peer as Owner Device
			// peer will automatically be removed from the staged list when set to be owner
			modStorestate.add( "--setowner" );
			modStorestate.add( peerId );

			returnValue |= DO_SYNCDB;
		}
		else {
			// There is already an owner device set for this study

			if ( isDeleted ) {
				// Check if the peer is a staged device and remove it
				if ( !stagedDevices.contains( peerId ) ) {
					// The peer is not a staged device
					return returnValue; //ie: DO_NOTHING
				}
				
				// remove peer from staged devices
				modStorestate.add( "--removedevice" );
				modStorestate.add( peerId );
			}
			else {
				// The peer should be added to staged devices
				if ( stagedDevices.contains( peerId ) ) {
					// The peer has already been added to staged devices
					return returnValue; //ie: DO_NOTHING
				}

				// add peer to staged devices
				modStorestate.add( "--setstaged" );
				modStorestate.add( peerId );
			}
		}
		
		if ( !storestate.modify( modStorestate ) ) {
			throw new IMException( "storestate.rec file could not be modified" );
		}

		return returnValue;
	}


	/**
	 * Checks if the STUDYDIR reference counter contains legacy references and convert legacy references to new style references
	 * Legacy references:                New style references:
	 *     FOLDER_<folder>                   <device>_FOLDER_<folder>
	 *     FNAME_<file>                      <device>_FNAME_<file>
	 *     PBRIUID_<report uid>              <device>_PBRIUID_<report uid>
	 * 
	 * @param refcounter     STUDYDIR reference counter
	 * @param storestate     storestate.rec content of the study
	 *
	 * @return true if and only if the reference counter has successfully been converted.
	 *
	 * @throws IOException if an I/O error occurs
	 */
	boolean checkAndFixRefcounter( ReferenceCounter refcounter, StoreState storestate ) throws IOException {
		if ( REFERENCE_COUNTER_COMMENT.equals( refcounter.getComment() ) ) {
			// Nothing to do, reference counter has already been converted
			return true;
		}

		if ( !refcounter.referrers().hasMoreElements() ) {
			// Nothing to do, either the reference counter does not exist or contain no referrers
			return true;
		}

		//String[] devices = storestate.getDotcomDevices();
		String ownerDevice = storestate.getAttribute( "OwnerDevice", "NULL" );
		if ( "NULL".equals( ownerDevice ) ) {
			logger.syslog( Syslog.LOG_ERR, "storestate.rec contains no valid ownerDevice" );
			return false;
		}	
		
		Vector<String> delReferences = new Vector<String>( 256, 256 );
		Vector<String> addReferences = new Vector<String>( 256, 256 );
		for ( Enumeration<String> e = refcounter.referrers(); e.hasMoreElements() ; ) {
			String ref = e.nextElement();
			if ( ref.startsWith( "FNAME_" ) || ref.startsWith( "FOLDER_" ) || ref.startsWith( "PBRIUID_" ) ) {
				delReferences.add( ref );
				addReferences.add( ownerDevice + "_" + ref );
			}
		}

		boolean isOK = refcounter.batch( null, addReferences, delReferences, REFERENCE_COUNTER_COMMENT );
		refcounter.load();
		return isOK;
	}

%><%
	long before = System.currentTimeMillis();
	
	IMSession is = new IMSession( request, response, out );
	String senderDevId = null;
	String studyUid = null;
	MultipartRequest mp = new MultipartRequest( is, IMEnvBase.getEnv( "TMP" ) );	
	boolean isForced = false;
	
	before = System.currentTimeMillis();

%><%--	// TODO: lock for avoid concurrent calls from two child servers --%><%
	
	int numTables = 0;
	try {
		studyUid = mp.getParameter( "cStudy" );
		numTables = Integer.parseInt( mp.getParameter( "cNumTables" ) );
		isForced = "yes".equals( mp.getParameter( "cForced" ) );
		
		Properties devPrp = DotcomToolbox.checkClient( is, "C", false );
		is.bRegNotNeeded = false;

		if ( devPrp != null )
			senderDevId = devPrp.getProperty( "ID" );
	}
	catch ( Throwable th ) {
		return;	
	}

	logger.syslog( Syslog.LOG_INFO, "syncStudy.jsp CALLED for study : " + studyUid + " Sender device : " + senderDevId + " ,numSources: " + mp.getParameter("cNumSources") );
	logger.syslog( Syslog.LOG_INFO, "syncStudy.jsp numTables: " + numTables + ( isForced ? " Forced " : "" ) );  
	
	if ( studyUid == null ) {
		// We might want to return error here
		return;
	}

	if ( StudyToolbox.isOrder( studyUid ) ) {
		logger.syslog( Syslog.LOG_INFO, "syncStudy.jsp was called for order (" + studyUid + "); ignoring" );
		out.print( "OK" );
		return;
	}


	IMScriptContext context = IMScriptContext.readJobContextFrom( request );
	
	boolean startAutoRetrieve = "true".equals( mp.getParameter( "cAutoRetrieve" ) );

	StoreState storestate = new StoreState( studyUid );
	storestate.load();
	boolean isRetrieved = "Staged".equals( storestate.getAttribute( "StageState" ) );
	boolean isDeleted = false;

	int syncTodo = DO_NOTHING; // Well, reference counters should always be handled
	if ( numTables != 0 ) {
		boolean hasValidValue = false; // If the study had been deleted, there would be no valid value
		for ( int i = 0; i < numTables ; i ++ ) {
			String uniqValues = mp.getParameter( "cUniqValues" + i );
			if ( null != uniqValues && !"".equals( uniqValues ) ) {
				
				Helper.DEBUG( "syncstudy.jsp valid unique value: " + uniqValues + " in table " + i );
				
				hasValidValue = true;
				break;
			}
			
			String values = mp.getParameter( "cValues" + i );
			if ( null != values && !"".equals( values ) ) {
				
				Helper.DEBUG( "syncstudy.jsp valid value: " + values + " in table " + i );
				
				hasValidValue = true;
				break;
			}

		}

		if( !hasValidValue ) {
			logger.syslog( Syslog.LOG_INFO, "syncStudy.jsp DELETED" );
			isDeleted = true;
		}
				
		syncTodo = checkAndFixStorestate( senderDevId, storestate, !hasValidValue );
	}

	ReferenceCounter refcounter = new ReferenceCounter( "STUDYDIR", studyUid );
	refcounter.load();
	
	if ( !checkAndFixRefcounter( refcounter, storestate ) ) {
		logger.syslog( Syslog.LOG_ERR, "Could not convert the STUDYDIR reference counter of study " + studyUid );
%><%--		// TODO: return error --%><%
		throw new RuntimeException( "Could not convert the STUDYDIR reference counter of study " + studyUid );
	}

	Vector<String> addReferences = new Vector<String>( 256, 256 );


	// First add the new entries to the reference counters/tables
	// Do it in two passes - add references first, then SQL table entries
	//    the second pass also removes unused SQL table entries and then unused references
	// Doing things in this order would cause the less problem if data become corrupted due to a broken connection

	// First pass : add new references, also collect references to be removed
	Map<String, Integer> toBeDeletedReferences = refcounter.copyReferrers();
 
	for ( int i = 0; i < numTables; i++ ) {
		String tableName = mp.getParameter( "cTableName" + i );

		// If this is a 'no object table' CWL server, don't add object table references 
		if ( DotcomConfig.selfDevice().isNoObjectTable() && "Dcobject".equals( tableName ) ) {
			continue;
		}

		String uniqAttr = mp.getParameter( "cUniqAttr" + i );
		String[] uniqvalues = mp.getParameter( "cUniqValues" + i ).split( "\n" );
		int tableRows = Integer.parseInt( mp.getParameter( "cTableRows" + i ) );

		if ( tableRows > 0 ) {
			for ( int j = 0; j < uniqvalues.length; j++ ) {
				if ( !refcounter.containsReferrer( senderDevId + "_" + uniqAttr + "_" + uniqvalues[ j ] ) ) {
					addReferences.add( senderDevId + "_" + uniqAttr + "_" + uniqvalues[ j ] );
					logger.syslog( Syslog.LOG_INFO, "syncstudy.jsp add reference " + senderDevId + "_" + uniqAttr + "_" + uniqvalues[ j ] + " to study " + studyUid );
				}
				// also remove reference from toBeDeletedReferences 
				toBeDeletedReferences.remove( senderDevId + "_" + uniqAttr + "_" + uniqvalues[ j ] );
			}
		}
	}

	if ( !addReferences.isEmpty() ) {
		// There are new references to be added
		logger.syslog( Syslog.LOG_INFO, "adding references to study " + studyUid );
		if ( !refcounter.batch( null, addReferences, null, REFERENCE_COUNTER_COMMENT ) ) {
			logger.syslog( Syslog.LOG_ERR, "Could not add references to study " + studyUid );
			throw new RuntimeException( "Could not add new references" );
		}
	}

	// Second pass : add/remove SQL entries
	
	
	if ( ( syncTodo & DO_SYNCDB ) == DO_SYNCDB ) {
		String logAction = "acquire";
		String logPNAME = null;
		String logMAINST = null;
		String logDerived = storestate.getAttribute( "Derived", "no" );
		String oldStudyStatus = null;
		
		for ( int i = 0; i < numTables; i++ ) {
			String tableName = mp.getParameter( "cTableName" + i );

			// If this is a 'no object table' CWL server, don't update object table entries
			if ( DotcomConfig.selfDevice().isNoObjectTable() && "Dcobject".equals( tableName ) ) {
				continue;
			}

			String uniqAttr = mp.getParameter( "cUniqAttr" + i );
			String[] uniqvalues = mp.getParameter( "cUniqValues" + i ).split( "\n" );
			String fields = mp.getParameter( "cFields" + i );
			String[] values = mp.getParameter( "cValues" + i ).split( "\n" );
			int tableRows = Integer.parseInt( mp.getParameter( "cTableRows" + i ) );

			//check if fields' database names exist
			//if not, exclude it form fields and all values 
			String[] actFieldVec = fields.split( "," );
			String newFields = "";
			boolean problem = false;
			HashSet<String> existingFieldSet = new HashSet<String>();
			DatabaseMetaData md = IMEnv.getDB().getMetaData();
			ResultSet metars = md.getColumns( null, null, tableName, "" );
			while ( metars.next() ) {
				existingFieldSet.add( metars.getString( "COLUMN_NAME" ) );
			}
			HashSet<Integer> invalidFields = new HashSet<Integer>();
			for ( int f = 0; f < actFieldVec.length; f++ ) {
				String actField = actFieldVec[ f ].trim();
				if ( !existingFieldSet.contains( actField ) ) {
					problem = true;
					invalidFields.add( new Integer( f ) );
					Helper.ERRORLOG( "The field '" + actField + "' cannot be found in table '" + tableName + "'" );
				}
				else {
					newFields += "," + actField;
				}
			}
			if ( problem ) {
				if ( newFields.length() > 0 ) {
					newFields = newFields.substring( 1 );
				}
				fields = newFields;
				if ( tableRows > 0 ) {
					for ( int j = 0; j < uniqvalues.length; j++ ) {
//						String[] valueVec = values[ j ].split( "," );
						String[] valueVec = Helper.splitNotQuoted( values[ j ], ",", '\'' );
						String newValues = "";
						for ( int fi = 0; fi < valueVec.length; fi++ ) {
							if ( !invalidFields.contains( new Integer( fi ) ) ) {
								newValues += "," + valueVec[ fi ];								
							}
						}
						if ( newValues.length() > 0 ) {
							newValues = newValues.substring( 1 );
						}
						values[ j ] = newValues;
					}
				}					
			}
			
			
			String query = "replace " + tableName + "(" + fields + ") values  (";

			StringBuffer uniqvaluesbuf = new StringBuffer();

			if ( tableRows > 0 ) {
				for ( int j = 0; j < uniqvalues.length; ) {
					// Run at most 20 MySQL queries in batch to avoid possible memory allocation problems
					Statement stat = IMEnv.getDB().createStatement();	
					for ( int k = 0; k < 20 && j < uniqvalues.length; j++, k++ ) {							
						stat.addBatch( ( new StringBuffer( query ) ).append( values[ j ] ).append( ")" ).toString() );
			
						Helper.DEBUG( "syncupdatequery: " + query + values[ j ] + ")" );
			
						if ( isForced ) {
							uniqvaluesbuf.append( ",'" ).append( Helper.sqlEncode( uniqvalues[ j ] ) ).append( "'" );
						}
						
						if ( "Dcobject".equals( tableName ) ) {
							//log all objects
							String actValues = values[ j ];
							String[] fieldVec = fields.split( "," );
//							String[] valueVec = actValues.split( "," );
							String[] valueVec = Helper.splitNotQuoted( actValues, ",", '\'' );
							String logFNAME = null;
							String logMODALITY = null;
							String logDevice = null; //don't know here
							String logSIZE = null;
							for ( int fi = 0; fi < fieldVec.length && fi < valueVec.length; fi++ ) {
								String actField = fieldVec[ fi ].trim();
								String actValue = valueVec[ fi ].trim();
								if ( "RETRIEVED".equals( actField ) ) { //not here!!
									if ( "yes".equals( stripQuotes( actValue ) ) ) {
										logAction = "retrieve";
									}
								}
								else if ( "FNAME".equals( actField ) ) {
									logFNAME = stripQuotes( actValue );
								}
								else if ( "OMODALITY".equals( actField ) ) {
									logMODALITY = stripQuotes( actValue );
								}
								else if ( "SIZE".equals( actField ) ) {
									logSIZE = actValue;
								}
							}
									
							OperLogEntry logEntry = new OperLogEntry( OperLogEntry.LEVEL.OBJECT, logAction );
							logEntry.setField( OperLogEntry.FIELD.USER, "system" );
							logEntry.setField( OperLogEntry.FIELD.FNAME, logFNAME );
							logEntry.setField( OperLogEntry.FIELD.STUDY, studyUid );
							logEntry.setField( OperLogEntry.FIELD.PNAME, logPNAME );
							logEntry.setField( OperLogEntry.FIELD.MODALITY, logMODALITY );
							logEntry.setField( OperLogEntry.FIELD.HOST, logDevice );
							logEntry.setField( OperLogEntry.FIELD.SIZE, logSIZE );
							logEntry.setField( OperLogEntry.FIELD.MESSAGE, "image" );
							logEntry.setField( OperLogEntry.FIELD.MAINST, logMAINST );
							logEntry.setField( OperLogEntry.FIELD.DERIVED, logDerived );
							logEntry.log( operLogger, Syslog.LOG_INFO );

//							operLogger.syslog( Syslog.LOG_INFO, "INFO OBJECT " + logAction + " system " + logFNAME + " " + studyUid + " " 
//										+ logPNAME + " " + logMODALITY + " "
//										+ logDevice + " " + logSIZE + " - - - - - image - " + logMAINST );
						}					
						else if( "Dcstudy".equals( tableName ) ) {
							//get old main status
							String statusQuery = "select MAINST from Dcstudy where STYIUID='" + studyUid + "'";
							Helper.DEBUG( "statusQuery:" + statusQuery );
							ResultSet rs = IMEnv.getDB().prepareStatement( statusQuery ).executeQuery();
							while( rs.next() ) { //there will be 1
								oldStudyStatus = rs.getString( 1 );
							}
						}
					}
					logger.syslog( Syslog.LOG_DEBUG, "syncstudy.jsp execute replace queries in batch for study " + studyUid + " [" + j + "]" );
					int[] affectedRows = stat.executeBatch();
					if ( "Dcstudy".equals( tableName ) ) {
						for ( int ri = 0; ri < affectedRows.length; ri++ ) { //in fact ri will be 0 only
							int row = j - affectedRows.length + ri;
							String actValues = values[ row ];
							String[] fieldVec = fields.split( "," );
//							String[] valueVec = actValues.split( "," );
							String[] valueVec = Helper.splitNotQuoted( actValues, ",", '\'' );
							String logMODALITY = null;
							String logDevice = null; //don't know here
							String logPID = null;
							String logACCNO = null;
							String logDATE = null;
							String logINSTNAME = null;
							String logSTYDESCR = null;
							for ( int fi = 0; fi < fieldVec.length && fi < valueVec.length; fi++ ) {
								String actField = fieldVec[ fi ].trim();
								String actValue = valueVec[ fi ].trim();
								if ( "RETRIEVED".equals( actField ) ) {
									if ( "yes".equals( stripQuotes( actValue ) ) ) {
										logAction = "retrieve";
									}
								}
								else if ( "PNAME".equals( actField ) ) {
									logPNAME = stripQuotes( actValue );
								}
								else if ( "MODALITY".equals( actField ) ) {
									logMODALITY = stripQuotes( actValue );
								}
								else if ( "PID".equals( actField ) ) {
									logPID = stripQuotes( actValue );
								}
								else if ( "ACCNO".equals( actField ) ) {
									logACCNO = stripQuotes( actValue );
								}
								else if ( "THEDATE".equals( actField ) ) {
									logDATE = stripQuotes( actValue );
								}
								else if ( "INSTNAME".equals( actField ) ) {
									logINSTNAME = stripQuotes( actValue );
								}
								else if ( "STYDESCR".equals( actField ) ) {
									logSTYDESCR = stripQuotes( actValue );
								}
								else if ( "MAINST".equals( actField ) ) {
									logMAINST = stripQuotes( actValue );
								}
								else if ( "FOLDER".equals( actField ) ) {
									String folder = stripQuotes( actValue );
									if (folder != null && folder.length() > 0) {
										//ensure that folder exists
										FolderToolbox.createFolder( is, folder, true );
									}
								}
							}
							
							if ( affectedRows[ ri ] == 1 ) {
								//log new studies only
								//file after system is not known
								//FNAME for Objects
								OperLogEntry logEntry = new OperLogEntry( OperLogEntry.LEVEL.STUDY, logAction );
								String logUSER = imagemed.dicom.RegisterDicomFile.checkAndReadUserFile( studyUid, "system" );
								logEntry.setField( OperLogEntry.FIELD.USER, logUSER );
								logEntry.setField( OperLogEntry.FIELD.STUDY, studyUid );
								logEntry.setField( OperLogEntry.FIELD.PNAME, logPNAME );
								logEntry.setField( OperLogEntry.FIELD.MODALITY, logMODALITY );
								logEntry.setField( OperLogEntry.FIELD.HOST, logDevice );
								logEntry.setField( OperLogEntry.FIELD.MESSAGE, "study_inserted" );
								logEntry.setField( OperLogEntry.FIELD.PID, logPID );
								logEntry.setField( OperLogEntry.FIELD.ACCNO, logACCNO );
								logEntry.setField( OperLogEntry.FIELD.STUDYDATE, logDATE );
								logEntry.setField( OperLogEntry.FIELD.INSTITUTION, logINSTNAME );
								logEntry.setField( OperLogEntry.FIELD.STUDYDESCR, logSTYDESCR );
								logEntry.setField( OperLogEntry.FIELD.STATUS, logMAINST );
								logEntry.setField( OperLogEntry.FIELD.DERIVED, logDerived );
								logEntry.log( operLogger, Syslog.LOG_INFO );

//								operLogger.syslog( Syslog.LOG_INFO, "INFO STUDY " + logAction + " system - " + studyUid + " " 
//										+ logPNAME + " " + logMODALITY + " "
//										+ logDevice + " - - - - - - study_inserted " + logPID 
//										+ " " + logACCNO + " " + logDATE + " " + logINSTNAME + " " + logSTYDESCR );
							}
							
							// log status change
							else if( logMAINST != null && oldStudyStatus != null && !oldStudyStatus.equals( logMAINST ) ) {
								OperLogEntry logEntry = new OperLogEntry( OperLogEntry.LEVEL.STUDY, "status" );
								logEntry.setField( OperLogEntry.FIELD.USER, "system" );
								logEntry.setField( OperLogEntry.FIELD.STUDY, studyUid );
								logEntry.setField( OperLogEntry.FIELD.PNAME, logPNAME );
								logEntry.setField( OperLogEntry.FIELD.MODALITY, logMODALITY );
								logEntry.setField( OperLogEntry.FIELD.HOST, logDevice );
								logEntry.setField( OperLogEntry.FIELD.MESSAGE, "Synchronized from HUB" );
								logEntry.setField( OperLogEntry.FIELD.PID, logPID );
								logEntry.setField( OperLogEntry.FIELD.ACCNO, logACCNO );
								logEntry.setField( OperLogEntry.FIELD.STUDYDATE, logDATE );
								logEntry.setField( OperLogEntry.FIELD.INSTITUTION, logINSTNAME );
								logEntry.setField( OperLogEntry.FIELD.STUDYDESCR, logSTYDESCR );
								logEntry.setField( OperLogEntry.FIELD.STATUS, logMAINST );
								logEntry.setField( OperLogEntry.FIELD.DERIVED, logDerived );
								logEntry.log( operLogger, Syslog.LOG_INFO );
							}
						}
					}
				}
			}

			//deleting may occur only if forced (full study/object/report) data is sent
			if ( isForced ) {
				
				String studyDB = IMEnvBase.getEnv( "REPORTTBL" ).equals( tableName ) ? "SRCSTUDY" : "STYIUID";
				
				if ( uniqvaluesbuf.length() > 0 ) { 
					query = "select " + uniqAttr + " from " + tableName + " where " + studyDB + "='" + studyUid + "' and " +  
						uniqAttr +	" not in (" + uniqvaluesbuf.substring( 1 ) + ")";	
				}
				else {
					query = "select " + uniqAttr + " from " + tableName + " where " + studyDB + "='" + studyUid + "'";
				}
			
				ResultSet rs = IMEnv.getDB().prepareStatement( query ).executeQuery();

				Vector<String> keys_to_delete = new Vector<String>();

				while ( rs.next() ) {
					keys_to_delete.add( rs.getString( 1 ) );
				}

				if ( keys_to_delete.size() > 0 ) {
					String delquery = "delete from " + tableName + " where " + studyDB + "='" + studyUid + 
						"' and " + uniqAttr + "='"; 

					for ( Iterator<String> it = keys_to_delete.iterator(); it.hasNext(); ) {
						// Run at most 20 MySQL queries in batch to avoid possible memory allocation problems
						Statement stat = IMEnv.getDB().createStatement();
						for ( int k = 0; k < 20 && it.hasNext(); k++ ) {
							String key = it.next();
							stat.addBatch( ( new StringBuffer( delquery ) ).append( key ).append( "'" ).toString() );
						}
					
						logger.syslog( Syslog.LOG_DEBUG, "syncstudy.jsp execute delete queries in batch for study " + studyUid + " [" + keys_to_delete.size() + "]" );
						stat.executeBatch();
					}
				}
			}
		} // end for i
	} // end if DO_SYNCDB

	
	// this solution works correctly to define if numsorcues is 1 or more
	// we don't really care if it's 2 or 3 or 4, as it's used to generate the yellow warning and more than 2 is really messed up
	// Update Number of Sources in the Dcstudy table
	// My parent doesn't count!
	int numSources = 0;
	String devs[] = storestate.getDotcomDevices();
	String parentId = DotcomConfig.selfParent();
	for ( int i = 0; i < devs.length; i++ ) {
		if ( !devs[ i ].equals( parentId ) ) {
			numSources++;
		}
	}

	//max( par, counted )
	String parNumSources = mp.getParameter( "cNumSources" );
	if ( parNumSources != null ) {
		try {
			int intParNumSources = Integer.parseInt( parNumSources );
			if ( intParNumSources > numSources ) {
				numSources = intParNumSources;
			}
		}
		catch ( NumberFormatException ex ) {
		}
	}

	StringBuilder sourceUpdate = new StringBuilder( 256 );
	sourceUpdate.append( "update " ).append( IMEnvBase.getEnv( "STUDYTBL" ) ).append( " set NUMSRC=" ).append( numSources );
	
	String ownerID = storestate.getAttribute( "OwnerDevice", "" );
	DotcomDevice ownerDev = DotcomConfig.getConfig().getDevice( ownerID );
	if ( ownerDev != null && !ownerDev.isDbOnly() ) {
		// Real hub
		sourceUpdate.append( ",OWNERHUB='" ).append( ownerID ).append( "'" );
	}
	sourceUpdate.append( " where STYIUID='" ).append( studyUid ).append( "'" );
	
	IMEnv.getDB().prepareStatement( sourceUpdate.toString() ).executeUpdate();

	if ( isDeleted ) {
		// The study just deleted from the leg
		if ( numSources <= 2 ) {
			// might be just cleared up by the multihub correction tool - delete 'in progress' flag, if any
			imagemed.dotcom.multihub.CorrectionTool.removeInProgressFlag( studyUid );
		}
	}

	// Deleted references which are to be removed
	// deleting may occur ONLY when forced (full tables) synch is sent!
	if ( !toBeDeletedReferences.isEmpty() && isForced ) {
		// Check and skip non sync. references

		Set<String> delReferences = toBeDeletedReferences.keySet();
		for ( Iterator<String> i = delReferences.iterator(); i.hasNext(); ) {
			String reference = i.next();
			if ( reference.startsWith( senderDevId + "_" ) ) {
				logger.syslog( Syslog.LOG_DEBUG, "reference to be removed: " + reference + " from study " + studyUid );
				continue;
			}

			// Reference from other devices or non sync reference - should not be removed from the reference counter
			i.remove();
		}

		if ( !delReferences.isEmpty() ) {
			if ( !refcounter.batch( null, null, delReferences, REFERENCE_COUNTER_COMMENT ) ) {
				logger.syslog( Syslog.LOG_ERR, "Could not remove references from study " + studyUid );
				throw new RuntimeException( "Could not remove references from STUDYDIR reference counter" );
			}
		}
	}

	if ( ( syncTodo & DO_CHANGE_OWNER) == DO_CHANGE_OWNER ) {
		// let the new owner re-synchronize this study
		String cmd = "dotcom.SyncFromChildTask -s \"" + studyUid + "\"";
		context.createChildTask( "SYNCHRONIZE_FROM_CHILD", cmd, "class" );
	}

	if ( startAutoRetrieve ) {
		// start an autoretrieve task for the study unless it's disabled
		if ( "yes".equals( IMEnv.getConfigProperty( IMEnvBase.getEnv( "CONF" ) + "/self.rec", "AutoRetrieve", "yes" ) ) ) {
			logger.syslog( Syslog.LOG_DEBUG, "Start autoretrieve task for study " + studyUid );
			PriorToolbox.startAutoretrieving( context, studyUid, storestate );
		}
	}

	if ( isRetrieved ) {
		// Study was retrieved from archive and the retrieve was initiated from the current server
		
		// Find and remove archive_..._get references if any
		for ( Enumeration<String> e = refcounter.referrers(); e.hasMoreElements(); ) {
			String ref = e.nextElement();
			if ( ref.startsWith( "archive_" ) && ref.endsWith( "_get" ) ) {
				refcounter.removeReferrer( ref );
			}
		}
		
		// Clear the staged state attribute of the storestate.rec file
		storestate.setMode( StoreState.EXISTING_ONLY );
		
		Vector<String> modStorestate = new Vector<String>();
		modStorestate.add( "--stagestate" );
		modStorestate.add( "" );

		storestate.modify( modStorestate );				
	}

	//synch to parent if it's CWL, too
	try {
		if ( DotcomConfig.isParentDbOnly() ) {
			if ( parentId != null ) {
				logger.syslog( Syslog.LOG_DEBUG, "syncStudy.jsp:: start synchronizing " + studyUid + " to " + parentId );
				StudySynchronizer.startTask( context, studyUid, isForced, true ); // force synchronization
			}
		}
	}
	catch ( Throwable th ) {
		logger.syslog( Syslog.LOG_ERR, "syncStudy.jsp EXCEPTION in synch to CWL parent: " + Helper.throwable2String( th ) );
		return;	
	}
	
	logger.syslog( Syslog.LOG_INFO, "syncStudy.jsp OK for study : " + studyUid + " [" + senderDevId + "] (" + ( System.currentTimeMillis() - before ) + " msec" );

%>OK
